@@ -3,7 +3,7 @@
 #
 #  (C) Paul Evans, 2021 -- leonerd@leonerd.org.uk
 
-package XS::Parse::Keyword 0.03;
+package XS::Parse::Keyword 0.04;
 
 use v5.14;
 use warnings;
@@ -56,6 +56,33 @@ installed by C<XS::Parse::Keyword> itself.
 The C<XSParseKeywordHooks> structure provides the following hook stages, which
 are invoked in the given order.
 
+=head2 flags
+
+The following flags are defined:
+
+=over 4
+
+=item C<XPK_FLAG_EXPR>
+
+The parse or build function is expected to return C<KEYWORD_PLUGIN_EXPR>.
+
+=item C<XPK_FLAG_STMT>
+
+The parse or build function is expected to return C<KEYWORD_PLUGIN_STMT>.
+
+These two flags are largely for the benefit of giving static information at
+registration time to assist static parsing or other related tasks to know what
+kind of grammatical element this keyword will produce.
+
+=item C<XPK_FLAG_AUTOSEMI>
+
+The syntax forms a complete statement, which should be followed by a statement
+separator semicolon (C<;>). This semicolon is optional at the end of a block.
+
+The semicolon, if present, will be consumed automatically.
+
+=back
+
 =head2 The C<permit> Stage
 
    const char *permit_hintkey;
@@ -108,7 +135,7 @@ result value.
 This is the most generic and powerful of the options, but requires the most
 amount of implementation work.
 
-   int (*build)(pTHX_ OP **out, XSParseKeywordPiece *args, size_t npieces, void *hookdata);
+   int (*build)(pTHX_ OP **out, XSParseKeywordPiece *args[], size_t nargs, void *hookdata);
 
 If C<parse> is not present, this is called instead after parsing a sequence of
 arguments, of types given by the I<pieces> field; which should be a zero-
@@ -137,6 +164,10 @@ function, using a C<union> type consisting of the following fields:
       CV *cv;
       SV *sv;
       int i;
+      struct {
+         SV *name;
+         SV *value;
+      } attr;
    } XSParseKeywordPiece;
 
 Which field is set depends on the type of the piece.
@@ -204,12 +235,65 @@ A bareword package name is expected, and passed as an SV containing a PV in
 the I<sv> field. A package name is similar to an identifier, except it permits
 double colons in the middle.
 
+=head2 XPK_LEXVARNAME
+
+I<atomic, emits sv.>
+
+   XPK_LEXVARNAME(kind)
+
+A lexical variable name is expected, and passed as an SV containing a PV in
+the I<sv> field. The C<kind> argument specifies what kinds of variable are
+permitted, and should be a bitmask of one or more bits from
+C<XPK_LEXVAR_SCALAR>, C<XPK_LEXVAR_ARRAY> and C<XPK_LEXVAR_HASH>. A convenient
+shortcut C<XPK_LEXVAR_ANY> permits all three.
+
+=head2 XPK_ATTRIBUTES
+
+I<atomic, emits i followed by more args.>
+
+A list of C<:>-prefixed attributes is expected, in the same format as sub or
+variable attributes. An optional leading C<:> indicates the presence of
+attributes, then one or more of them are parsed. Attributes may be optionally
+separated by additional C<:>s, but this is not required.
+
+Each attribute is expected to be an identifier name, followed by an optional
+value wrapped in parentheses. Whitespace is B<NOT> permitted between the name
+and value, as per standard Perl parsing rules.
+
+   :attrname
+   :attrname(value)
+
+The I<i> field indicates how many attributes were found. That number of
+additional arguments are then passed, each containing two SVs in the
+I<attr.name> and I<attr.value> fields. This number may be zero.
+
+It is not an error for there to be no attributes present, or for the optional
+colon to be missing. In this case I<i> will be set to zero.
+
+=head2 XPK_VSTRING
+
+I<atomic, can probe, emits sv.>
+
+A version string is expected, of the form C<v1.234> including the leading C<v>
+character. It is passed as a L<version> SV object in the I<sv> field.
+
+=head2 XPK_VSTRING_OPT
+
+Identical to C<XPK_VSTRING> except it is optional; if no version string is
+found then I<sv> is set to C<NULL>.
+
 =head2 XPK_COLON
 
-I<atomic, emits nothing.>
+I<atomic, can probe, emits nothing.>
 
 A literal colon character (C<:>) is expected. No argument value is passed.
 
+=head2 XPK_EQUALS
+
+I<atomic, can probe, emits nothing.>
+
+A literal equals character (C<=>) is expected. No argument value is passed.
+
 =head2 XPK_STRING
 
 I<atomic, can probe, emits nothing.>
@@ -277,6 +361,16 @@ which is passed in the I<i> field.
 
    XPK_TAG(value)
 
+=head2 XPK_COMMALIST
+
+I<structural, emits i.>
+
+A structural type which expects to find one or more repeats of its contained
+pieces, separated by literal comma (C<,>) characters. This is somewhat similar
+to C<XPK_REPEATED>, except that it needs at least one copy, needs commas
+between its items, but does not require that the first contained piece support
+probe (the comma itself is sufficient to indicate a repeat).
+
 =head2 XPK_PARENSCOPE
 
 I<structural, emits nothing.>
