@@ -3,7 +3,7 @@
 #
 #  (C) Paul Evans, 2021 -- leonerd@leonerd.org.uk
 
-package XS::Parse::Keyword::Builder 0.03;
+package XS::Parse::Keyword::Builder 0.04;
 
 use v5.14;
 use warnings;
@@ -132,6 +132,12 @@ struct XSParseKeywordPieceType {
   } u;
 };
 
+enum {
+  XPK_FLAG_EXPR     = (1<<0),
+  XPK_FLAG_STMT     = (1<<1),
+  XPK_FLAG_AUTOSEMI = (1<<2),
+};
+
 enum {
   /* skip zero */
   XS_PARSE_KEYWORD_BLOCK = 1,     /* op */
@@ -142,12 +148,16 @@ enum {
   /* TODO: XS_PARSE_KEYWORD_FULLEXPR = 6 */
   XS_PARSE_KEYWORD_IDENT = 7,     /* sv */
   XS_PARSE_KEYWORD_PACKAGENAME,   /* sv */
+  XS_PARSE_KEYWORD_LEXVARNAME,    /* sv */
+  XS_PARSE_KEYWORD_ATTRS,         /* attr.name + attr.val */
+  XS_PARSE_KEYWORD_VSTRING,       /* sv */
   XS_PARSE_KEYWORD_LITERALCHAR = 0x10,
   XS_PARSE_KEYWORD_LITERALSTR,
   XS_PARSE_KEYWORD_OPTIONAL = 0x20,
   XS_PARSE_KEYWORD_REPEATED,
   XS_PARSE_KEYWORD_CHOICE,
   XS_PARSE_KEYWORD_TAGGEDCHOICE,
+  XS_PARSE_KEYWORD_SEPARATEDLIST,
   XS_PARSE_KEYWORD_FAILURE = 0x2f,
   XS_PARSE_KEYWORD_PARENSCOPE = 0x30, /* (...) */
   XS_PARSE_KEYWORD_BRACKETSCOPE,      /* [...] */
@@ -155,6 +165,17 @@ enum {
   XS_PARSE_KEYWORD_CHEVRONSCOPE,      /* <...> */
 };
 
+enum {
+  XPK_LEXVAR_SCALAR = (1<<0),
+  XPK_LEXVAR_ARRAY  = (1<<1),
+  XPK_LEXVAR_HASH   = (1<<2),
+  XPK_LEXVAR_ANY    = XPK_LEXVAR_SCALAR|XPK_LEXVAR_ARRAY|XPK_LEXVAR_HASH,
+};
+
+enum {
+  XS_PARSE_KEYWORD_TYPEFLAG_OPT = (1<<16),
+};
+
 #define XPK_BLOCK    {.type = XS_PARSE_KEYWORD_BLOCK}
 #define XPK_ANONSUB  {.type = XS_PARSE_KEYWORD_ANONSUB}
 #define XPK_TERMEXPR {.type = XS_PARSE_KEYWORD_TERMEXPR}
@@ -162,44 +183,56 @@ enum {
 #define XPK_IDENT    {.type = XS_PARSE_KEYWORD_IDENT}
 #define XPK_PACKAGENAME {.type = XS_PARSE_KEYWORD_PACKAGENAME}
 
-#define XPK_COLON {.type = XS_PARSE_KEYWORD_LITERALCHAR, .u.c = ':'}
+#define XPK_LEXVARNAME(kind) {.type = XS_PARSE_KEYWORD_LEXVARNAME, .u.c = kind}
+
+#define XPK_ATTRIBUTES {.type = XS_PARSE_KEYWORD_ATTRS}
+
+#define XPK_VSTRING     {.type = XS_PARSE_KEYWORD_VSTRING}
+#define XPK_VSTRING_OPT {.type = XS_PARSE_KEYWORD_VSTRING|XS_PARSE_KEYWORD_TYPEFLAG_OPT}
+
+#define XPK_COLON  {.type = XS_PARSE_KEYWORD_LITERALCHAR, .u.c = ':'}
+#define XPK_EQUALS {.type = XS_PARSE_KEYWORD_LITERALCHAR, .u.c = '='}
 
 #define XPK_STRING(s) {.type = XS_PARSE_KEYWORD_LITERALSTR, .u.str = (const char *)s}
 
 /* First piece of these must be something probe-able */
 #define XPK_OPTIONAL(...) \
-  {.type = XS_PARSE_KEYWORD_OPTIONAL, .u.pieces = (const struct XSParseKeywordPieceType []){ __VA_ARGS__, 0 }}
+  {.type = XS_PARSE_KEYWORD_OPTIONAL, .u.pieces = (const struct XSParseKeywordPieceType []){ __VA_ARGS__, {0} }}
 #define XPK_REPEATED(...) \
-  {.type = XS_PARSE_KEYWORD_REPEATED, .u.pieces = (const struct XSParseKeywordPieceType []){ __VA_ARGS__, 0 }}
+  {.type = XS_PARSE_KEYWORD_REPEATED, .u.pieces = (const struct XSParseKeywordPieceType []){ __VA_ARGS__, {0} }}
 /* Every piece must be probeable */
 #define XPK_CHOICE(...) \
-  {.type = XS_PARSE_KEYWORD_CHOICE, .u.pieces = (const struct XSParseKeywordPieceType []){ __VA_ARGS__, 0 }}
+  {.type = XS_PARSE_KEYWORD_CHOICE, .u.pieces = (const struct XSParseKeywordPieceType []){ __VA_ARGS__, {0} }}
 /* Every piece must be probeable, and followed by XPK_TAG */
 #define XPK_TAGGEDCHOICE(...) \
-  {.type = XS_PARSE_KEYWORD_TAGGEDCHOICE, .u.pieces = (const struct XSParseKeywordPieceType []){ __VA_ARGS__, 0, 0 }}
+  {.type = XS_PARSE_KEYWORD_TAGGEDCHOICE, .u.pieces = (const struct XSParseKeywordPieceType []){ __VA_ARGS__, {0}, {0} }}
 #define XPK_TAG(val) \
   {.type = val}
+#define XPK_COMMALIST(...) \
+  {.type = XS_PARSE_KEYWORD_SEPARATEDLIST, .u.pieces = (const struct XSParseKeywordPieceType []){ \
+      {.type = XS_PARSE_KEYWORD_LITERALCHAR, .u.c = ','}, __VA_ARGS__, {0}}}
 
 #define XPK_FAILURE(s) {.type = XS_PARSE_KEYWORD_FAILURE, .u.str = (const char *)s}
 
 #define XPK_PARENSCOPE(...) \
-  {.type = XS_PARSE_KEYWORD_PARENSCOPE, .u.pieces = (const struct XSParseKeywordPieceType []){ __VA_ARGS__, 0 }}
+  {.type = XS_PARSE_KEYWORD_PARENSCOPE, .u.pieces = (const struct XSParseKeywordPieceType []){ __VA_ARGS__, {0} }}
 #define XPK_BRACKETSCOPE(...) \
-  {.type = XS_PARSE_KEYWORD_BRACKETSCOPE, .u.pieces = (const struct XSParseKeywordPieceType []){ __VA_ARGS__, 0 }}
+  {.type = XS_PARSE_KEYWORD_BRACKETSCOPE, .u.pieces = (const struct XSParseKeywordPieceType []){ __VA_ARGS__, {0} }}
 #define XPK_BRACESCOPE(...) \
-  {.type = XS_PARSE_KEYWORD_BRACESCOPE, .u.pieces = (const struct XSParseKeywordPieceType []){ __VA_ARGS__, 0 }}
+  {.type = XS_PARSE_KEYWORD_BRACESCOPE, .u.pieces = (const struct XSParseKeywordPieceType []){ __VA_ARGS__, {0} }}
 #define XPK_CHEVRONSCOPE(...) \
-  {.type = XS_PARSE_KEYWORD_CHEVRONSCOPE, .u.pieces = (const struct XSParseKeywordPieceType []){ __VA_ARGS__, 0 }}
+  {.type = XS_PARSE_KEYWORD_CHEVRONSCOPE, .u.pieces = (const struct XSParseKeywordPieceType []){ __VA_ARGS__, {0} }}
 
 typedef union {
   OP *op;
   CV *cv;
   SV *sv;
   int i;
+  struct { SV *name; SV *value; } attr;
 } XSParseKeywordPiece;
 
 struct XSParseKeywordHooks {
-  U32 flags; /* not used yet but reserves the space without breaking ABI later */
+  U32 flags;
 
   /* used by build1 */
   struct XSParseKeywordPieceType piece1;
@@ -214,7 +247,7 @@ struct XSParseKeywordHooks {
 
   /* These are alternatives; the first one defined is used */
   int (*parse)(pTHX_ OP **opp, void *hookdata);
-  int (*build)(pTHX_ OP **out, XSParseKeywordPiece *pieces, size_t npieces, void *hookdata);
+  int (*build)(pTHX_ OP **out, XSParseKeywordPiece *args[], size_t nargs, void *hookdata);
   int (*build1)(pTHX_ OP **out, XSParseKeywordPiece arg0, void *hookdata);
 };
 
@@ -238,13 +271,26 @@ static void S_boot_xs_parse_keyword(pTHX_ double ver) {
   svp = hv_fetchs(PL_modglobal, "XS::Parse::Keyword/ABIVERSION", 0);
   if(!svp)
     croak("XS::Parse::Keyword ABI version missing");
-  int abi_version = SvIV(*svp);
-  if(abi_version != XSPARSEKEYWORD_ABI_VERSION)
+  int abi_ver = SvIV(*svp);
+  if(abi_ver != XSPARSEKEYWORD_ABI_VERSION)
     croak("XS::Parse::Keyword ABI version mismatch - library provides %d, compiled for %d",
-        abi_version, XSPARSEKEYWORD_ABI_VERSION);
+        abi_ver, XSPARSEKEYWORD_ABI_VERSION);
+
+  if((svp = hv_fetchs(PL_modglobal, "XS::Parse::Keyword/ABIVERSION_MIN", 0))) {
+      abi_ver = SvIV(*svp);
+      if(abi_ver > XSPARSEKEYWORD_ABI_VERSION)
+        croak("XS::Parse::Keyword ABI version mismatch - library supports >= %d, compiled for %d",
+            abi_ver, XSPARSEKEYWORD_ABI_VERSION);
+
+      svp = hv_fetchs(PL_modglobal, "XS::Parse::Keyword/ABIVERSION_MAX", 0);
+      abi_ver = SvIV(*svp);
+      if(abi_ver < 1)
+        croak("XS::Parse::Keyword ABI version mismatch - runtime requires at least version 1, library only supports %d",
+            abi_ver);
+  }
 
   register_xs_parse_keyword_func = INT2PTR(void (*)(pTHX_ const char *, const struct XSParseKeywordHooks *, void *),
-      SvUV(*hv_fetchs(PL_modglobal, "XS::Parse::Keyword/register()", 0)));
+      SvUV(*hv_fetchs(PL_modglobal, "XS::Parse::Keyword/register()@1", 0)));
 }
 
 #endif
